/* Copyright (C) 2000, 2002, 2003, 2004, 2005, 2008 Free Software Foundation,
   Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307 USA.  */

#ifndef _LINUX_NIOS2_SYSDEP_H
#define _LINUX_NIOS2_SYSDEP_H 1

#include <sysdeps/unix/sysdep.h>

/* For Linux we can use the system call table in the header file
        /usr/include/asm/unistd.h
   of the kernel.  But these symbols do not follow the SYS_* syntax
   so we have to redefine the `SYS_ify' macro here.  */
#undef SYS_ify
#define SYS_ify(syscall_name)   __NR_##syscall_name

#ifdef __ASSEMBLER__

#define ASM_TYPE_DIRECTIVE(name,typearg) .type name,%##typearg;
#define ASM_SIZE_DIRECTIVE(name) .size name,.-name

#define ENTRY(name)                                                           \
  ASM_GLOBAL_DIRECTIVE C_SYMBOL_NAME(name);                                   \
  ASM_TYPE_DIRECTIVE (C_SYMBOL_NAME(name),function)                           \
  C_LABEL(name)

#undef  END
#define END(name)                                                             \
  ASM_SIZE_DIRECTIVE(name)

#define SYSCALL_ERROR_LABEL __syscall_error

#undef PSEUDO
#define PSEUDO(name, syscall_name, args) \
  ENTRY (name)                           \
    DO_CALL (syscall_name, args)         \
    bne r7, zero, SYSCALL_ERROR_LABEL;   \

#undef PSEUDO_END
#define PSEUDO_END(name) \
  SYSCALL_ERROR_HANDLER  \
  END (name)

#undef PSEUDO_NOERRNO
#define PSEUDO_NOERRNO(name, syscall_name, args) \
  ENTRY (name)                                   \
    DO_CALL (syscall_name, args)

#undef PSEUDO_END_NOERRNO
#define PSEUDO_END_NOERRNO(name) \
  END (name)

#undef ret_NOERRNO
#define ret_NOERRNO ret

#undef DO_CALL
#define DO_CALL(syscall_name, args) \
    DOARGS_##args                   \
    movi r2, SYS_ify(syscall_name);  \
    trap;

#if defined(__PIC__) || defined(PIC)

#define SYSCALL_ERROR_HANDLER \
SYSCALL_ERROR_LABEL:          \
  subi sp, sp, 12;            \
  stw r22, 8(sp);             \
  stw r2, 4(sp);              \
  stw ra, 0(sp);              \
  nextpc r22;                 \
1: movhi r8, %hiadj(_GLOBAL_OFFSET_TABLE_ - 1b); \
  addi r8, r8, %lo(_GLOBAL_OFFSET_TABLE_ - 1b); \
  add r22, r22, r8;           \
  ldw r3, %call(__errno_location)(r22); \
  callr r3;                   \
  ldw ra, 0(sp);              \
  ldw r10, 4(sp);             \
  ldw r22, 8(sp);             \
  addi sp, sp, 12;            \
  stw r10, 0(r2);             \
  movi r2, -1;                \
  ret;

#else

#define SYSCALL_ERROR_HANDLER \
SYSCALL_ERROR_LABEL:          \
  subi sp, sp, 8;             \
  stw r2, 4(sp);              \
  stw ra, 0(sp);              \
  call __errno_location;      \
  ldw ra, 0(sp);              \
  ldw r10, 4(sp);             \
  addi sp, sp, 8;             \
  stw r10, 0(r2);             \
  movi r2, -1;                \
  ret;

#endif

#define DOARGS_0 /* nothing */
#define DOARGS_1 /* nothing */
#define DOARGS_2 /* nothing */
#define DOARGS_3 /* nothing */
#define DOARGS_4 /* nothing */
#define DOARGS_5 ldw r8, 0(sp);
#define DOARGS_6 ldw r9, 4(sp); ldw r8, 0(sp);

/* The function has to return the error code.  */
#undef  PSEUDO_ERRVAL
#define PSEUDO_ERRVAL(name, syscall_name, args) \
  ENTRY (name)                                  \
    DO_CALL (syscall_name, args)

#undef  PSEUDO_END_ERRVAL
#define PSEUDO_END_ERRVAL(name) \
  END (name)

#define ret_ERRVAL ret

#else /* __ASSEMBLER__ */

#include <asm/unistd.h>

/* Define a macro which expands into the inline wrapper code for a system
   call.  */
#undef INLINE_SYSCALL
#define INLINE_SYSCALL(name, nr, args...)                               \
  ({ INTERNAL_SYSCALL_DECL(err);					\
     unsigned int result_var = INTERNAL_SYSCALL (name, err, nr, args);	\
     if ( INTERNAL_SYSCALL_ERROR_P (result_var, err) )			\
       {								\
	 __set_errno (INTERNAL_SYSCALL_ERRNO (result_var, err));	\
	 result_var = -1L;						\
       }								\
     (int) result_var; })

#undef INTERNAL_SYSCALL_DECL
#define INTERNAL_SYSCALL_DECL(err) unsigned int err

#undef INTERNAL_SYSCALL_ERROR_P
#define INTERNAL_SYSCALL_ERROR_P(val, err)   ((unsigned int) (err))

#undef INTERNAL_SYSCALL_ERRNO
#define INTERNAL_SYSCALL_ERRNO(val, err)     (val)

#undef INTERNAL_SYSCALL_RAW
#define INTERNAL_SYSCALL_RAW(name, err, nr, args...)            \
  ({ unsigned int _sys_result;                                  \
     {                                                          \
       register int _r2 asm ("r2") = (int)(name);               \
       register int _err asm ("r7");                            \
       LOAD_ARGS_##nr (args)                                    \
       asm volatile ("trap"                                     \
                     : "=r" (_r2), "=r" (_err)                  \
                     : ASM_ARGS_##nr                            \
                     : __SYSCALL_CLOBBERS);                     \
       _sys_result = _r2;                                       \
       err = _err;                                              \
     }                                                          \
     (int) _sys_result; })

#undef INTERNAL_SYSCALL
#define INTERNAL_SYSCALL(name, err, nr, args...) \
	INTERNAL_SYSCALL_RAW(SYS_ify(name), err, nr, args)

#undef INTERNAL_SYSCALL_NCS
#define INTERNAL_SYSCALL_NCS(number, err, nr, args...) \
	INTERNAL_SYSCALL_RAW(number, err, nr, args)

#define LOAD_ARGS_0()
#define ASM_ARGS_0                  "r" (_r2)
#define LOAD_ARGS_1(a1)                         \
  register int _r4 asm ("r4") = (int) (a1);     \
  LOAD_ARGS_0 ()
#define ASM_ARGS_1      ASM_ARGS_0, "r" (_r4)
#define LOAD_ARGS_2(a1, a2)                     \
  register int _r5 asm ("r5") = (int) (a2);     \
  LOAD_ARGS_1 (a1)
#define ASM_ARGS_2      ASM_ARGS_1, "r" (_r5)
#define LOAD_ARGS_3(a1, a2, a3)                 \
  register int _r6 asm ("r6") = (int) (a3);     \
  LOAD_ARGS_2 (a1, a2)
#define ASM_ARGS_3      ASM_ARGS_2, "r" (_r6)
#define LOAD_ARGS_4(a1, a2, a3, a4)             \
  register int _r7 asm ("r7") = (int) (a4);     \
  LOAD_ARGS_3 (a1, a2, a3)
#define ASM_ARGS_4      ASM_ARGS_3, "r" (_r7)
#define LOAD_ARGS_5(a1, a2, a3, a4, a5)         \
  register int _r8 asm ("r8") = (int) (a5);     \
  LOAD_ARGS_4 (a1, a2, a3, a4)
#define ASM_ARGS_5      ASM_ARGS_4, "r" (_r8)
#define LOAD_ARGS_6(a1, a2, a3, a4, a5, a6)     \
  register int _r9 asm ("r9") = (int) (a6);     \
  LOAD_ARGS_5 (a1, a2, a3, a4, a5)
#define ASM_ARGS_6      ASM_ARGS_5, "r" (_r9)


#define __SYSCALL_CLOBBERS "r1", "r3", "r10", "r11", "r12", "r13", "r14", \
                           "r15", "r29", "memory"

#endif /* __ASSEMBLER__ */

#endif /* linux/nios2/sysdep.h */
